;*****************************************************************************
;
;  trgrep : Search for a pattern in file(s) or stdin if no
;           file name is specified.
;
;  usage : trgrep [-c] [-n] [-q] [-v] PATTERN [FILE]...
;
;          -c  match count only  (ignored if -q)
;          -n  line numbers      (ignored if -c or -q)
;          -q  set exit code only(0 = match, 1 = no match, 2 = error)
;          -v  select non-matching lines
;
;       If no FILE is specified then the input is read from stdin.
;       FILE may contain wildcard characters(*?).
;
;*****************************************************************************


	format pe console
	entry  __start


BUFFER_SIZE  equ  8192

;**********************************************************
section ".code" code readable executable
;**********************************************************
include "windows.inc"
include "..\ctypedef.inc"
include "cmdline.inc"
include "itoa.inc"
include "..\AsmRegEx.inc"

__start:

	; set file name for error message
	mov	dword[fname_str], stdout_str
	mov	dword[fname_len], stdout_len

	; get stdout handle
       @sys.get_stdhandle  STD_OUTPUT_HANDLE

	cmp	eax, INVALID_HANDLE_VALUE
	mov	[stdout_handle], eax
	jz   .err_fopen

	; get pointer to command line string
       @sys.get_commandline
	mov	esi, eax
	call	parse_cmdline	; setup "argc", "argv" and "cmdln_options"
	test	eax, eax
	jz   .error_msg_exit	; no/invalid args or invalid switch

	; allocate input buffer from stack
	mov	[buf_end], esp
	sub	esp, BUFFER_SIZE
	xor	eax, eax
	mov	[buf_ptr],  esp
	mov	[line_ptr], esp
	mov	[line_len], eax
	mov	edi, esp
	mov	ecx, (BUFFER_SIZE shr 2)
	rep  stosd

	xor	ecx, ecx
	inc	byte[cur_arg]	; index of next arg
	dec	ecx
	mov	esi, [argv]	; first arg has to be the pattern

 .pattern_len:
	inc	ecx
	cmp	byte[esi+ecx], 0
	jnz  .pattern_len
	mov	[pattern_ptr], esi
	mov	[pattern_len], ecx
	; compile pattern
	; esi == pointer to 0-terminated pattern
	push	esi
	push	lpREGEX_T
	call	regcomp
	test	eax, eax
	jz   .error_msg_exit

 .get_input:
	; test if file name(s) are specified
	cmp	dword[argc], 1
	ja   .next_arg

	; no file name --> read from stdin
	inc	byte[stdin_flag]

	; set file name for error message
	mov	dword[fname_str], stdin_str
	mov	dword[fname_len], stdin_len

	; get stdin handle
       @sys.get_stdhandle  STD_INPUT_HANDLE

	cmp	eax, INVALID_HANDLE_VALUE
	mov	[file_handle], eax
	jz   .err_fopen
	jmp  .print_file_name
	;--------------------

 .next_arg:
	movzx	eax, byte[cur_arg]
	xor	ebx, ebx
	cmp	eax, [argc]
	jz   .done
	shl	eax, 2
	inc	byte[cur_arg]
	mov	esi, [argv+eax]
	xor	ecx, ecx
	xor	eax, eax
	mov	[fname_str], esi
	inc	ebx

	; test if file name contains a wildcard (*?) character
	; and get the length of the file name
 .has_wildcard:
	lodsb
	test	eax, eax
	jz   .end_file_name
	inc	ecx
	cmp	eax, "?"
	jz   .wild_card
	cmp	eax, "*"
	jnz  .has_wildcard

  .wild_card:
	mov	[wild_card], bl
	jmp  .has_wildcard
	;--------------------

 .end_file_name:
	test	[wild_card], bl
	mov	[fname_len], ecx
	jnz  .init_find_file

  .open_file:
	mov	ebx, [fname_str]

       @sys.fopen

	cmp	eax, INVALID_HANDLE_VALUE
	mov	[file_handle], eax
	jz   .err_fopen

  .print_file_name:
	test	dword[cmdln_options], Q_OPTION
	jnz  .fill_buffer
	mov	ecx, [fname_str]
	mov	edx, [fname_len]
	call	sys.print_string
	; print ":\n"
	mov	ecx, colon_lf_str
	mov	edx, colon_lf_len
	call	sys.print_string

 .fill_buffer:
	; fill input buffer:
	;       read BUFFER_SIZE bytes from "file_handle" to buffer
	;       pointed to by "buf_ptr"

	mov	ebx, [file_handle]
	mov	ecx, [buf_ptr]
	mov	edx, BUFFER_SIZE

       @sys.fread

	test	eax, eax
	jz   .err_fread
	mov	[buf_len], ecx	; number of bytes read(size of data in buffer)
	test	ecx, ecx
	jz   .end_of_input

 .get_line:
	; get a line from the input buffer
	xor	ecx, ecx
	mov	edx, [buf_len]
	mov	eax, [line_ptr]
	mov	ebx, [line_len]
	cmp	ecx, edx
	jz   .end_of_input

	; add the length of the previous line (initially 0) to the start
	; of the previous line (initially start of buffer) to get the
	; start of the next line
	add	eax, ebx
	; subtract the length of the previous line (initially 0) to get
	; the number of bytes left in the buffer (size of data to process)
	sub	edx, ebx
	; set the length of the current line to 0
	mov	[line_len], ecx
	; esi <- start of current line
	mov	esi, eax
	; set the size of the data in the buffer
	mov	[buf_len],  edx
	; set the start of the current line
	mov	[line_ptr], eax
	; test if the buffer has been exhausted
	test	edx, edx
	; eax <- 0
	mov	eax, ecx
	jle  .refill_buffer

  .line_len:
	lodsb
	; length of line + 1
	inc	ecx
	cmp	eax, LF
	jz   .line_end
	; size of data in buffer - 1
	dec	edx
	jnz  .line_len
	; test if the current line spans across the buffer boundary
	cmp	esi, [buf_end]
	jae  .refill_buffer
	; set buffer data size to 0
	mov	[buf_len], edx

   .line_end:
	mov	esi, [line_ptr]
	mov	[line_len], ecx
	inc	dword[line_count]
	xor	ebx, ebx
	; save first character of next line and zero terminate current line
	; (regexec expects the search data to be 0-terminated)
	movzx	eax, byte[esi+ecx]
	push	eax
	mov	[esi+ecx], bl

	push	esi
	push	lpREGEX_T
	call	regexec
	; eax == 0 if not found, != 0 if found

	mov	esi, [line_ptr]
	mov	ecx, [line_len]
	pop	ebx
	; test if there is a match within the current line
	test	eax, eax
	; restore first character of next line
	mov	[esi+ecx], bl
	jz   .not_found

   .found:
	test	dword[cmdln_options], V_OPTION
	jnz  .get_line

    .check_match_options:
	xor	ebx, ebx
	inc	dword[match_count]
	inc	dword[have_match]
	test	dword[cmdln_options], Q_OPTION
	jnz  .done	; terminate program with exit code 0 (= match found)
	; test if matching count only is requested
	test	dword[cmdln_options], C_OPTION
	jnz  .get_line	; match has been counted --> continue search
	; test if line numbers are requested
	test	dword[cmdln_options], N_OPTION
	jz   .display_matchline
	; line numbers are requested
	mov	eax, [line_count]
	mov	edi, itoa_buffer
       @itoa
	; ebx == length of ASCII representation of line number
	; edi == pointer to next free byte after the converted line number
	mov	byte[edi], ":"
	inc	ebx		; + ":"
	mov	ecx, itoa_buffer
	mov	edx, ebx
	call	sys.print_string

  .display_matchline:
	mov	ecx, [line_ptr]
	mov	edx, [line_len]
	call	sys.print_string
	jmp  .get_line
	;--------------------

 .not_found:
	; there is no match whithin the current line
	; test if non-matching lines are requested
	test	dword[cmdln_options], V_OPTION
	jz   .get_line		; !found && !V_OPTION
	; non-matching lines are requested --> we have a match
	jmp  .check_match_options
	;--------------------

 .refill_buffer:
	; end of buffer reached
	; ecx == length of current line

	; either the current line spans accross the buffer boundary
	; or     the end of input has been reached

	; test if the current line is longer than the buffer
	cmp	ecx, BUFFER_SIZE
	jae  .err_line_too_long

	; copy partial line to start of buffer
	mov	edi, [buf_ptr]
	mov	edx, ecx	; length of partial line
	mov	ebx, ecx	; length of partial line
	mov	esi, [line_ptr] ; start  of partial line
	shr	ecx, 2
	mov	[line_ptr], edi ; line_ptr <- buf_ptr
	and	edx, 3
	rep  movsd
	mov	ecx, edx
	mov	[line_len], ebx ; length of partial line
	mov	eax, BUFFER_SIZE
	rep  movsb
	; partial line moved to start of buffer

	sub	eax, ebx	; BUFFER_SIZE - length of partial line
	mov	ebx, [file_handle]
	mov	ecx, edi
	mov	edx, eax

       @sys.fread

	test	eax, eax
	mov	[buf_len], ecx	; number of bytes read
	jz   .err_fread
	test	ecx, ecx
	jz   .end_of_input
	mov	edx, ecx	; number of bytes read
	mov	esi, edi	; current buffer position
	mov	ecx, [line_len] ; length of partial line
	xor	eax, eax
	add	dword[buf_len], ecx
	jmp  .line_len
	;--------------------

 .end_of_input:
	test	byte [stdin_flag], 1
	mov	ebx, [file_handle]
	jnz  .reset
	; ebx == file handle
       @sys.fclose

  .reset:
	xor	eax, eax
	mov	ebx, [buf_ptr]
	mov	[line_count], eax
	mov	[line_len],   eax
	mov	[buf_len],    eax
	mov	[line_ptr],   ebx

	test	dword[cmdln_options], C_OPTION
	jz   .more_input
	; print count of matching lines
	mov	edi, itoa_buffer
	mov	eax, [match_count]
	test	eax, eax
	jz   .more_input
       @itoa
	; ebx == length of ASCII representation of line number
	; edi == pointer to next free byte after the converted line number
	mov	byte[edi], LF
	inc	ebx	; + LF
	mov	ecx, itoa_buffer
	mov	edx, ebx
	call	sys.print_string
	xor	eax, eax

 .more_input:
	test	byte[stdin_flag],1
	jnz   .done
	test	byte[wild_card], 1
	mov	[match_count], eax
	jz    .next_arg

  .next_file:
	push	w32_find_data
	push	dword[w32_find_handle]
	call	[FindNextFileA]
	test	eax, eax
	jz   .no_more_files

   .init_file:
	xor	ecx, ecx
	test	dword[w32_find_data.dwFileAttributes], FILE_ATTRIBUTE_DIRECTORY
	jnz  .next_file
	dec	ecx
	mov	esi, w32_find_data.cFileName

   .file_name_len:
	inc	ecx
	cmp	byte[esi+ecx], 0
	jnz  .file_name_len
	mov	[fname_len], ecx
	mov	[fname_str], esi
	jmp  .open_file
	;--------------------

 .no_more_files:
	; eax == 0
	mov	[wild_card], al
	push	dword[w32_find_handle]
	call	[FindClose]
	jmp  .next_arg
	;---------------------

 .init_find_file:
	; file name contains a wildcard (*?)
	push	w32_find_data
	push	dword[fname_str]
	call	[FindFirstFileA]
	cmp	eax, INVALID_HANDLE_VALUE
	mov	[w32_find_handle], eax
	jz   .no_such_file
	jmp  .init_file
	;--------------------

 .done:
	cmp	dword[have_match], 0
	setz	bl

  .exit:
       @sys.exit
	;====================

.no_such_file:
	mov	edx, err_no_such_file_len
	mov	ecx, err_no_such_file_str
	call	sys.print_string
	mov	edx, [fname_len]
	mov	ecx, [fname_str]
	mov	byte[edx+ecx], LF
	call	sys.print_string
	xor	eax, eax
	mov	edx, [fname_len]
	mov	ecx, [fname_str]
	mov	[edx+ecx],  al
	mov	[wild_card],al
	jmp  .next_arg
	;--------------------

.err_line_too_long:
	mov	edx, err_line_len_len
	mov	ecx, err_line_len_str
	jmp  .error_msg_exit
	;--------------------

.err_fread:
	test	dword[cmdln_options], Q_OPTION
	jnz  .error_exit
	mov	edx, err_fread_len
	mov	ecx, err_fread_str
	jmp  .print_msg_fname
	;--------------------

.err_fopen:
	test	dword[cmdln_options], Q_OPTION
	jnz  .error_exit
	mov	edx, err_fopen_len
	mov	ecx, err_fopen_str

 .print_msg_fname:
	call	sys.print_string
	mov	edx, [fname_len]
	mov	ecx, [fname_str]

.error_msg_exit:
	test	dword[cmdln_options], Q_OPTION
	jnz  .error_exit
	call	sys.print_string

 .error_exit:
	mov	ebx, 2
       @sys.exit

;**********************************************************
section ".data" data readable writeable
;**********************************************************

err_cmdln_too_long_str	db "command line too long",10
err_cmdln_too_long_len	 = $ - err_cmdln_too_long_str
err_inv_option_str	db "invalid option",10
err_inv_option_len	 = $ - err_inv_option_str
err_inv_arg_str 	db "invalid argument",10
err_inv_arg_len 	 = $ - err_inv_arg_str
err_broken_string_str	db "unbalanced double quotes",10
err_broken_string_len	 = $ - err_broken_string_str
err_too_many_args_str	db "too many arguments",10
err_too_many_args_len	 = $ - err_too_many_args_str
err_no_such_file_str	db "No files matching: "
err_no_such_file_len	 = $ - err_no_such_file_str
err_fopen_str		db "cannot open file: "
err_fopen_len		 = $ - err_fopen_str
err_fread_str		db "cannot read file: "
err_fread_len		 = $ - err_fread_str
err_line_len_str	db "input line exceeds buffer size",10
err_line_len_len	 = $ - err_line_len_str

; error messages of regcomp()
include "..\AsmRegEr.inc"

usage_str		db 10
			db "Usage: trgrep [-c] [-n] [-q] [-v] PATTERN [FILE]...",10
			db 10
			db 9,"-c  match count only  (ignored if -q)",10
			db 9,"-n  line numbers only (ignored if -q or -c)",10
			db 9,"-q  set exit code only(0 = match, 1 = no match, 2 = error)",10
			db 9,"-v  select non-matching lines",10
			db 10
			db " If no FILE is specified then the input is read from stdin.",10
			db " FILE may contain wildcard characters(*?).",10
usage_len		 = $ - usage_str

stdin_str		db "stdin"
stdin_len		 = $ - stdin_str
stdout_str		db "stdout"
stdout_len		 = $ - stdout_str

colon_lf_str		db ":",LF
colon_lf_len		 = $ - colon_lf_str

; character classification table
include "..\ctypemap.inc"

;**********************************************************
; uninitialized data
;**********************************************************
pattern_ptr	rd 1
pattern_len	rd 1
fname_str	rd 1
fname_len	rd 1

buf_ptr 	rd 1
buf_len 	rd 1
buf_end 	rd 1
line_ptr	rd 1
line_len	rd 1

line_count	rd 1
match_count	rd 1
have_match	rd 1

lpREGEX_T	REGEX_T

w32_find_data	WIN32_FIND_DATA
w32_find_handle rd 1
file_handle	rd 1
stdout_handle	rd 1


cur_arg 	rb 1
wild_card	rb 1
stdin_flag	rb 1
itoa_buffer	rb 13

ALIGN	4
cmdln_options	rd 1
argc		rd 1
argv		rd MAX_ARGV
args		rb MAX_ARGS

;**********************************************************
section ".idata" import data readable writeable
;**********************************************************
  dd 0,0,0,rva kernel_name,rva kernel_table
  dd 0,0,0,0,0

  kernel_table:
    GetCommandLineA  dd rva _GetCommandLineA
    GetStdHandle     dd rva _GetStdHandle
    FindFirstFileA   dd rva _FindFirstFileA
    FindNextFileA    dd rva _FindNextFileA
    FindClose	     dd rva _FindClose
    CreateFileA      dd rva _CreateFileA
    ReadFile	     dd rva _ReadFile
    WriteFile	     dd rva _WriteFile
    CloseHandle      dd rva _CloseHandle
    ExitProcess      dd rva _ExitProcess
    dd 0

  kernel_name db 'KERNEL32.DLL',0

  _GetCommandLineA dw 0
    db 'GetCommandLineA',0
  _GetStdHandle dw 0
    db 'GetStdHandle',0
  _FindFirstFileA dw 0
    db 'FindFirstFileA',0
  _FindNextFileA dw 0
    db 'FindNextFileA',0
  _FindClose dw 0
    db 'FindClose',0
  _CreateFileA dw 0
    db 'CreateFileA',0
  _ReadFile dw 0
    db 'ReadFile',0
  _WriteFile dw 0
    db 'WriteFile',0
  _CloseHandle dw 0
    db 'CloseHandle',0
  _ExitProcess dw 0
    db 'ExitProcess',0
